package me.km.snuviscript;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import me.hammerle.snuviscript.code.SnuviUtils;
import me.km.databank.DataBank;
import me.km.playerbank.PlayerManager;

public class ScriptBank implements IScriptBank {
    private final DataBank databank;
    private final PreparedStatement getVar;

    public ScriptBank(DataBank databank, PlayerManager manager) {
        this.databank = databank;
        databank.execute("CREATE TABLE IF NOT EXISTS scriptdata ("
                + "id INT NOT NULL AUTO_INCREMENT PRIMARY KEY, "
                + "player_id INT NOT NULL, "
                + "var VARCHAR(20) NOT NULL, "
                + "value VARCHAR(256) NOT NULL, "
                + "UNIQUE INDEX (player_id,var), "
                + "FOREIGN KEY (player_id) REFERENCES minecraft.players(id) ON DELETE RESTRICT);");

        getVar = databank.prepareStatement("SELECT value from scriptdata WHERE player_id = ? and var = ?;");
    }

    @Override
    public void setVar(String value, String var, int playerId) throws Exception {
        try(PreparedStatement setVar = databank.prepareStatement(
                "INSERT INTO scriptdata (player_id,var,value) SELECT ?, ?, ? ON DUPLICATE KEY UPDATE scriptdata.value=?;")) {
            setVar.setInt(1, playerId);
            setVar.setString(2, var);
            setVar.setString(3, value);
            setVar.setString(4, value);
            setVar.executeUpdate();
        }
    }

    @Override
    public void deleteVar(String var, int playerId) throws Exception {
        try(PreparedStatement deleteVar = databank.prepareStatement(
                "DELETE FROM scriptdata WHERE player_id = ? AND var = ?;")) {
            deleteVar.setInt(1, playerId);
            deleteVar.setString(2, var);
            deleteVar.executeUpdate();
        }
    }

    @Override
    public Object getVar(String var, int playerId, Object error) throws Exception {
        getVar.setInt(1, playerId);
        getVar.setString(2, var);
        try(ResultSet rs = getVar.executeQuery()) {
            if(rs.next()) {
                return SnuviUtils.convert(rs.getString(1));
            }
            return error;
        }
    }

    @Override
    public Object getVar(String var, int playerId) throws Exception {
        return getVar(var, playerId, null);
    }
}
